@@ -1,4 +1,4 @@
-import subprocess, os, struct, re, datetime
+import struct, re, datetime
 import hglib
 from hglib import error, util, templates, merge, context
 
@@ -46,8 +46,16 @@ class hgclient(object):
         self._args = [hglib.HGPATH, 'serve', '--cmdserver', 'pipe',
                 '--config', 'ui.interactive=True']
         if path:
-            self._args += ['-R', path]
+            # perhaps path shouldn't be a unicode string, but accepted for
+            # backward compatibility.
+            if isinstance(path, str):
+                # py2: bytes + bytes, py3: unicode + unicode
+                self._args += ['-R' + path]
+            else:
+                # py2: (ascii) bytes + unicode, py3: bytes + bytes
+                self._args += [b('-R') + path]
         if configs:
+            # don't use "--config=<value>" form for hg 1.9 compatibility
             for config in configs:
                 self._args += ['--config', config]
         self._env = {'HGPLAIN': '1'}
@@ -59,9 +67,52 @@ class hgclient(object):
         # include the hidden changesets if True
         self.hidden = None
 
+        self._cbout = None
+        self._cberr = None
+        self._cbprompt = None
+
         if connect:
             self.open()
 
+        self._protocoltracefn = None
+
+    def setcbout(self, cbout):
+        """
+        cbout is a function that will be called with the stdout data of
+        the command as it runs. Call with None to stop getting call backs.
+        """
+        self._cbout = cbout
+
+    def setcberr(self, cberr):
+        """
+        cberr is a function that will be called with the stderr data of
+        the command as it runs.Call with None to stop getting call backs.
+        """
+        self._cberr = cberr
+
+    def setcbprompt(self, cbprompt):
+        """
+        cbprompt is used to reply to prompts by the server
+        It receives the max number of bytes to return and the
+        contents of stdout received so far.
+
+        Call with None to stop getting call backs.
+
+        cbprompt is never called from merge() or import_()
+        which already handle the prompt.
+        """
+        self._cbprompt = cbprompt
+
+    def setprotocoltrace(self, tracefn=None):
+        """
+        if tracefn is None no trace calls will be made.
+        Otherwise tracefn is call as tracefn( direction, channel, data )
+        direction is 'r' for read from server and 'w' for write to server
+        channel is always None when direction is 'w'
+        and the channel-identified when the direction is 'r'
+        """
+        self._protocoltracefn = tracefn
+
     def __enter__(self):
         if self.server is None:
             self.open()
@@ -96,7 +147,11 @@ class hgclient(object):
     def _readchannel(self):
         data = self.server.stdout.read(hgclient.outputfmtsize)
         if not data:
-            raise error.ServerError()
+            ret, serr = self._close()
+            if ret != 0:
+                raise error.ServerError('server exited with status %d: %s'
+                                        % (ret, serr.strip()))
+            raise error.ResponseError('no response received from server')
         channel, length = struct.unpack(hgclient.outputfmt, data)
         if channel in b('IL'):
             return channel, length
@@ -119,6 +174,8 @@ class hgclient(object):
 
     def runcommand(self, args, inchannels, outchannels):
         def writeblock(data):
+            if self._protocoltracefn is not None:
+                self._protocoltracefn('w', None, data)
             self.server.stdin.write(struct.pack(self.inputfmt, len(data)))
             self.server.stdin.write(data)
             self.server.stdin.flush()
@@ -131,6 +188,8 @@ class hgclient(object):
 
         while True:
             channel, data = self._readchannel()
+            if self._protocoltracefn is not None:
+                self._protocoltracefn('r', channel, data)
 
             # input channels
             if channel in inchannels:
@@ -143,6 +202,7 @@ class hgclient(object):
                 return struct.unpack(hgclient.retfmt, data)[0]
             # a channel that we don't know and can't ignore
             elif channel.isupper():
+                self.close()
                 raise error.ResponseError(
                     "unexpected data on required channel '%s'" % channel)
             # optional channel
@@ -164,14 +224,32 @@ class hgclient(object):
         It receives the max number of bytes to return
         """
         out, err = BytesIO(), BytesIO()
-        outchannels = {b('o') : out.write, b('e') : err.write}
+        outchannels = {}
+        if self._cbout is None:
+            outchannels[b('o')] = out.write
+        else:
+            def out_handler(data):
+                out.write(data)
+                self._cbout(data)
+            outchannels[b('o')] = out_handler
+        if self._cberr is None:
+            outchannels[b('e')] = err.write
+        else:
+            def err_handler(data):
+                err.write(data)
+                self._cberr(data)
+            outchannels[b('e')] = err_handler
 
         inchannels = {}
+        if prompt is None:
+            prompt = self._cbprompt
         if prompt is not None:
             def func(size):
                 reply = prompt(size, out.getvalue())
                 return reply
             inchannels[b('L')] = func
+        else:
+            inchannels[b('L')] = lambda _: ''
         if input is not None:
             inchannels[b('I')] = input
 
@@ -192,7 +270,15 @@ class hgclient(object):
         self.server = util.popen(self._args, self._env)
         try:
             self._readhello()
+        except error.ResponseError:
+            if self.server is not None:
+                self._close()
+            raise
         except error.ServerError:
+            if self.server is None:
+                # server is already closed, hopefully the ServerError
+                # we got has enough information.
+                raise
             ret, serr = self._close()
             raise error.ServerError('server exited with status %d: %s'
                                     % (ret, serr.strip()))
@@ -539,6 +625,11 @@ class hgclient(object):
                           u=updaterev, r=revrange)
         self.rawcommand(args)
 
+    def init(self, dest, ssh=None, remotecmd=None, insecure=False):
+        args = util.cmdbuilder('init', dest, e=ssh, remotecmd=remotecmd,
+                               insecure=insecure)
+        self.rawcommand(args)
+
     def commit(self, message=None, logfile=None, addremove=False,
                closebranch=False, date=None, user=None, include=None,
                exclude=None, amend=False):
@@ -571,7 +662,8 @@ class hgclient(object):
         m = re.search(b(r'^committed changeset (\d+):([0-9a-f]+)'), out,
                       re.MULTILINE)
         if not m:
-            raise ValueError('revision and node not found in hg output: %r' % out)
+            raise ValueError('revision and node not found in hg output: %r'
+                             % out)
         rev, node = m.groups()
         return int(rev), node
 
@@ -597,7 +689,8 @@ class hgclient(object):
 
         conf = []
         if showsource:
-            out = util.skiplines(out, b('read config from: '))
+            out = util.skiplines(out, (b('read config from: '),
+                                       b('set config by: ')))
             for line in out.splitlines():
                 m = re.match(b(r"(.+?:(?:\d+:)?) (.*)"), line)
                 t = splitline(m.group(2))
@@ -776,9 +869,8 @@ class hgclient(object):
 
         args = cmdbuilder(b('grep'), all=all, a=text, f=follow, i=ignorecase,
                           l=fileswithmatches, n=line, u=user, d=date,
-                          I=include, X=exclude, hidden=self.hidden,
+                          I=include, X=exclude, hidden=self.hidden, print0=True,
                           *[pattern] + files)
-        args.append(b('-0'))
 
         def eh(ret, out, err):
             if ret != 1:
@@ -1060,13 +1152,14 @@ class hgclient(object):
 
         """
         # we can't really use --preview since merge doesn't support --template
-        args = cmdbuilder(b('merge'), r=rev, f=force, t=tool)
+        args = cmdbuilder(b('merge'), r=rev, f=force, t=tool,
+                          y=(cb is merge.handlers.noninteractive))
 
         prompt = None
         if cb is merge.handlers.abort:
             prompt = cb
         elif cb is merge.handlers.noninteractive:
-            args.append(b('-y'))
+            pass
         else:
             prompt = lambda size, output: cb(output) + b('\n')
 
@@ -1397,9 +1490,7 @@ class hgclient(object):
         args = cmdbuilder(b('status'), rev=rev, change=change, A=all,
                           m=modified, a=added, r=removed, d=deleted, c=clean,
                           u=unknown, i=ignored, C=copies, S=subrepos, I=include,
-                          X=exclude, hidden=self.hidden)
-
-        args.append(b('-0'))
+                          X=exclude, hidden=self.hidden, print0=True)
 
         out = self.rawcommand(args)
         l = []
@@ -1528,7 +1619,7 @@ class hgclient(object):
             elif name == b('branch'):
                 pass
             elif name == b('commit'):
-                value = value == b('(clean)')
+                value = b('(clean)') in value
             elif name == b('update'):
                 if value == b('(current)'):
                     value = 0
@@ -1631,7 +1722,10 @@ class hgclient(object):
         return self._version
 
     def __getitem__(self, changeid):
-        return context.changectx(self, changeid)
+        try:
+            return context.changectx(self, changeid)
+        except ValueError as e:
+            raise KeyError(*e.args)
 
     def __contains__(self, changeid):
         """
